function [xGN, k, rho, X, Alphas] = tikgn( fun, b, x0, lam, niter, mnflag, xbar)
%TIKGN Gauss-Newton method regularized by Tikhonov in standard form
%   [x, k, rho, X, alpha] = tikgn( fun, b, x0, lam, niter, mnflag, xbar)
%   computes the regularized solution to the nonlinear least-square
%   problem min||fun(x)-b||, i.e.
%   min { ||fun(x)-b||^2 + lam^2||x-xbar||^2 }.
%   where
%      fun    - function and Jacobian: [f,J]=fun(x)
%      b      - data vector
%      x0     - initial point
%      lam    - regularization parameter
%      niter  - max number of iterations
%      mnflag - 1 minimal-norm solution, 0 standard solution
%      xbar   - a priori estimate of solution (only if mnflag=1)
%   and
%      x      - solution
%      k      - number of iterations performed
%      rho    - solution residual
%      X      - solution at each iteration
%      alpha  - step-lenghts for damped Gauss-Newton method

%   F. Pes and G. Rodriguez
%   University of Cagliari, Italy

% Last revised July 24, 2020

if nargin<7, xbar = zeros(size(x0)); end
if nargin<6, mnflag = 1; end
if nargin<5, niter = 40; end

if lam<0, error('Illegal regularization parameter lambda.'), end
if mnflag==0 && ~(isequal(xbar,zeros(size(x0)))), error('xbar only if mnflag=1.'), end

minpowAG = -30;		% alpha > 2^minpowAG for Armijo-Goldstein

n = size(x0,1);
m = size(b,1);
r = min(m,n);

X = zeros(n,niter);
Alphas = zeros(niter,1);

normx0 = norm(x0);
X(:,1) = x0;
rGN = fun(X(:,1))-b;
rho = norm(rGN);

k = 0;
flag = 1;
while flag
	k = k+1;

	[fGN,jacGN] = fun(X(:,k));
	rGN = fGN-b;

	[U,S,V] = svd(jacGN);
	if min(m,n)==1
		d = S(1,1);
	else
		d = diag(S);
	end

	beta = U(:,1:r)'*rGN;
	d = d(1:r);
	if mnflag
		z = V'*(X(:,k)-xbar);
		y = -(d.*beta+lam^2*z(1:r))./(d.^2+lam^2);
	else
		y = -(d.*beta)./(d.^2+lam^2);
	end
	step = V(:,1:r)*y;

	% damped
	alpha = 2;
	flagd = 1;
	while flagd
		alpha = alpha/2;
		x1 = X(:,k) + alpha * step;
		rho1 = norm(fun(x1)-b);
		% Armijo-Goldstein principle
		alphas = norm(alpha*step);
		flagd = (rho^2 - rho1^2) < (1/2)*alpha*norm(jacGN*step)^2;
		flagalpha = alphas > pow2(minpowAG);
		flagd = flagd && flagalpha;
	end
	Alphas(k) = alpha;
	%     x1 = X(:,k) + alpha*step;
	if mnflag
		x1 = x1 - V(:,r+1:n)*z(r+1:n);
	end
	X(:,k+1) = x1;
	normx = norm(X(:,k+1));

	rho = norm(fun(x1)-b);

	diff = norm(X(:,k+1)-X(:,k)); % norm(x_k+1 - x_k)

	flagGN1 = (diff > 1e-8*max(1,normx)); % tol = 1e-8
	flagGN2 = (normx < 1e8*normx0);
	flagGN3 = (k < niter);
	%[flagGN1 flagGN2 flagGN3 flagalpha]
	flag = flagGN1 && flagGN2 && flagGN3 && flagalpha;
end

xGN = X(:,k+1);
X = X(:,1:k+1);
Alphas = Alphas(1:k);

